#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>

#define KEYDIR "ActiveState"

#if defined(_WIN32)
#include <direct.h>
#include <shlobj.h>
#include <windows.h>
typedef BOOL (WINAPI *PFNSHGetSpecialFolderPath)(HWND, char*, int, BOOL);
#else
#include <pwd.h>
#include <unistd.h>
#define _S_IFDIR S_IFDIR
#endif

#include "License.h"


#if defined(_WIN32)
static char *
_GetAppDir(void)
{
    char *path;
    PFNSHGetSpecialFolderPath pfn;
    HMODULE module = LoadLibrary("shell32.dll");
    if (!module)
        return NULL;

    pfn = (PFNSHGetSpecialFolderPath)GetProcAddress(module, "SHGetSpecialFolderPathA");
    if (pfn == NULL)
        goto Fail;

    path = (char*)malloc(MAX_PATH+1);
    if (!path)
        goto Fail;

    if (pfn(NULL, path, CSIDL_APPDATA, TRUE)) {
        FreeLibrary(module);
        return path;
    }

    free(path);

Fail:
    FreeLibrary(module);
    return NULL;
}

static BOOL
_VerifyDir(char **dir, char **err)
{
    struct stat statbuf;
    if (stat(*dir, &statbuf)) {
        *err = (char*)malloc(2048);
        sprintf(*err, "Path '%s' does not exist.", *dir);
        return FALSE;
    }

    if ((statbuf.st_mode & _S_IFDIR) == 0) {
        *err = (char*)malloc(2048);
        sprintf(*err, "'%s' is not a directory.", *dir);
        return FALSE;
    }

    return TRUE;
}
#endif

static BOOL
licGetHomeDir(char **dir, int *len, char **err)
{
#if !defined(_WIN32)
    struct passwd *p;
    char *env1 = NULL;
#else
    char *env1 = _GetAppDir();
    char *env2 = NULL;
#endif
    struct stat statbuf;

    if (!env1)
        env1 = getenv("HOME");

#if !defined(_WIN32)
    if (!env1) {
        p = getpwuid(getuid());
        env1 = p->pw_dir;
    }
    *dir = (char *)malloc(strlen(env1) + sizeof KEYDIR + 3);
    *len = sprintf(*dir, "%s/.%s/", env1, KEYDIR);

    /* create it if it doesn't exist yet */
    if (stat(*dir, &statbuf) || ((statbuf.st_mode & _S_IFDIR) == 0)) {
        if (mkdir(*dir, 0755)) {
            *err = (char*)malloc(2048);
            sprintf(*err, "Could not create directory '%s'.", *dir);
            return FALSE;
        }
    }
#else
    if (!env1)
        env1 = getenv("USERPROFILE");

    if (!env1) {
        env1 = getenv("HOMEDRIVE");
        env2 = getenv("HOMEPATH");
    }

    if (!env1)
        env1 = getenv("windir");

    if (!env1) {
        *err = strdup("Couldn't find HOME / USERPROFILE / HOMEDRIVE&HOMEPATH / windir");
        return FALSE;
    }

    if (!env2)
        env2 = "";

    /* make a copy of env1 and append env2 */
    *len = strlen(env1) + strlen(env2);
    /* allocate some more bytes to append '\ActiveState\' later */
    *dir = (char*)malloc(*len+20);
    strcat(strcpy(*dir, env1), env2);

    /* normalize path seperators to backslashes */
    for (env1=*dir; *env1; ++env1) {
        if (*env1 == '/')
            *env1 = '\\';
    }

    /* Win32 _stat() doesn't like trailing backslashes, except for "x:\" */
    if (*len > 3 && (*dir)[*len-1] == '\\')
        (*dir)[--(*len)] = '\0';

    /* verify that we have a directory */
    if (!_VerifyDir(dir, err))
        return FALSE;

    /* make sure we have a trailing backslash */
    if ((*dir)[*len-1] != '\\') {
        (*dir)[(*len)++] = '\\';
        (*dir)[*len] = '\0';
    }

    /* append our own subdirectory */
    strcat(*dir, KEYDIR);
    *len += 11;

    /* create it if it doesn't exist yet */
    if (stat(*dir, &statbuf) || ((statbuf.st_mode & _S_IFDIR) == 0)) {
        if (mkdir(*dir)) {
            *err = (char*)malloc(2048);
            sprintf(*err, "Could not create directory '%s'.", *dir);
            return FALSE;
        }
    }

    /* verify that we have a directory */
    if (!_VerifyDir(dir, err))
        return FALSE;

    /* append trailing backslash */
    (*dir)[(*len)++] = '\\';
    (*dir)[*len] = '\0';
#endif

    return TRUE;
}

static int
licCompareFeatures(const char *feature1, const char *feature2)
{
    if (!feature1)
        feature1 = "";
    if (!feature2)
        feature2 = "";

    while (1) {
        char ch1 = *feature1++;
        char ch2 = *feature2++;

        if (ch1 == '|') ch1 = '\0';
        if (ch2 == '|') ch2 = '\0';
        if (ch1 < ch2) return -1;
        if (ch1 > ch2) return +1;
        if (ch1 == '\0') return 0;
    }
}
